@@ -1,6 +1,7 @@
 //! Transactional filesystem traits and data structures. Unstable.
 
 use std::collections::HashSet;
+use std::convert::TryFrom;
 use std::io;
 use std::ops::{Deref, DerefMut};
 
@@ -14,7 +15,7 @@ use tokio_util::io::StreamReader;
 
 use tc_error::*;
 use tc_value::Value;
-use tcgeneric::{Id, PathSegment, TCBoxTryFuture};
+use tcgeneric::{Id, PathSegment, TCBoxTryFuture, TCTryStream};
 
 use super::{Transaction, TxnId};
 
@@ -32,12 +33,8 @@ pub trait BlockData: de::FromStream<Context = ()> + Clone + Send + Sync + 'stati
     where
         Self: en::ToStream<'en>,
     {
-        let mut data = destream_json::en::encode(self).map_err(TCError::internal)?;
         let mut hasher = Sha256::default();
-        while let Some(chunk) = data.try_next().map_err(TCError::internal).await? {
-            hasher.update(&chunk);
-        }
-
+        hash_chunks(&mut hasher, self).await?;
         let digest = hasher.finalize();
         Ok(Bytes::from(digest.to_vec()))
     }
@@ -226,12 +223,21 @@ pub trait Dir: Store + Sized + 'static {
     async fn create_dir_tmp(&self, txn_id: TxnId) -> TCResult<Self>;
 
     /// Create a new [`Self::File`].
-    async fn create_file<C: Send>(&self, txn_id: TxnId, name: Id, class: C) -> TCResult<Self::File>
+    async fn create_file<F: TryFrom<Self::File, Error = TCError>, C: Send>(
+        &self,
+        txn_id: TxnId,
+        name: Id,
+        class: C,
+    ) -> TCResult<F>
     where
         Self::FileClass: From<C>;
 
     /// Create a new [`Self::File`] with a new unique ID.
-    async fn create_file_tmp<C: Send>(&self, txn_id: TxnId, class: C) -> TCResult<Self::File>
+    async fn create_file_tmp<F: TryFrom<Self::File, Error = TCError>, C: Send>(
+        &self,
+        txn_id: TxnId,
+        class: C,
+    ) -> TCResult<F>
     where
         Self::FileClass: From<C>;
 
@@ -239,7 +245,11 @@ pub trait Dir: Store + Sized + 'static {
     async fn get_dir(&self, txn_id: &TxnId, name: &PathSegment) -> TCResult<Option<Self>>;
 
     /// Get a [`Self::File`] in this `Dir`.
-    async fn get_file(&self, txn_id: &TxnId, name: &Id) -> TCResult<Option<Self::File>>;
+    async fn get_file<F: TryFrom<Self::File, Error = TCError>>(
+        &self,
+        txn_id: &TxnId,
+        name: &Id,
+    ) -> TCResult<Option<F>>;
 }
 
 /// Defines how to load a persistent data structure from the filesystem.
@@ -260,3 +270,52 @@ pub trait Persist<D: Dir, T: Transaction<D>>: Sized {
 pub trait Restore<D: Dir, T: Transaction<D>>: Sized {
     async fn restore(&self, backup: &Self, txn_id: TxnId) -> TCResult<()>;
 }
+
+/// Defines how to copy a base state from another instance, possibly a view.
+#[async_trait]
+pub trait CopyFrom<D: Dir, T: Transaction<D>, Instance>: Persist<D, T> {
+    /// Copy a new instance of `Self` from an existing instance.
+    async fn copy_from(
+        instance: Instance,
+        store: <Self as Persist<D, T>>::Store,
+        txn_id: TxnId,
+    ) -> TCResult<Self>
+    where
+        Instance: 'async_trait;
+}
+
+/// Defines a standard hash for a persistent state.
+#[async_trait]
+pub trait Hash<'en> {
+    type Item: en::IntoStream<'en> + Send + 'en;
+
+    async fn hash_hex(&'en self, txn_id: TxnId) -> TCResult<String> {
+        self.hash(txn_id).map_ok(|hash| hex::encode(hash)).await
+    }
+
+    async fn hash(&'en self, txn_id: TxnId) -> TCResult<Bytes> {
+        let mut data = self.hashable(txn_id).await?;
+
+        let mut hasher = Sha256::default();
+        while let Some(item) = data.try_next().await? {
+            hash_chunks(&mut hasher, item).await?;
+        }
+
+        let digest = hasher.finalize();
+        Ok(Bytes::from(digest.to_vec()))
+    }
+
+    async fn hashable(&'en self, txn_id: TxnId) -> TCResult<TCTryStream<'en, Self::Item>>;
+}
+
+async fn hash_chunks<'en, T: en::IntoStream<'en> + 'en>(
+    hasher: &mut Sha256,
+    data: T,
+) -> TCResult<()> {
+    let mut data = destream_json::en::encode(data).map_err(TCError::internal)?;
+    while let Some(chunk) = data.try_next().map_err(TCError::internal).await? {
+        hasher.update(&chunk);
+    }
+
+    Ok(())
+}
